import contextlib
import json
from typing import TYPE_CHECKING, Any

from any_agent import AgentFramework
from any_agent.tracing.processors.base import TracingProcessor

if TYPE_CHECKING:
    from any_agent.tracing.trace import AgentSpan


class OpenAITracingProcessor(TracingProcessor):
    """Processor for OpenAI agent trace."""

    def _get_agent_framework(self) -> AgentFramework:
        return AgentFramework.OPENAI

    def _extract_llm_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        attributes = span.attributes
        if not attributes:
            msg = "Span attributes are empty"
            raise ValueError(msg)
        span_info = {}

        input_key = "llm.input_messages.1.message.content"
        if input_key in attributes:
            span_info["input"] = attributes[input_key]

        output_content = None
        for key in [
            "llm.output_messages.0.message.content",
            "llm.output_messages.0.message.contents.0.message_content.text",
        ]:
            if key in attributes:
                output_content = attributes[key]
                break

        if output_content:
            span_info["output"] = output_content

        return span_info

    def _extract_tool_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        attributes = span.attributes
        if not attributes:
            msg = "Span attributes are empty"
            raise ValueError(msg)
        tool_name = attributes.get("tool.name", "Unknown tool")
        tool_output = attributes.get("output.value", "")

        span_info = {
            "tool_name": tool_name,
            "input": attributes.get("input.value", ""),
            "output": tool_output,
        }

        with contextlib.suppress(json.JSONDecodeError):
            span_info["input"] = json.loads(span_info["input"])

        return span_info

    def _extract_agent_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract information from an AGENT span."""
        span_info: dict[str, Any] = {
            "type": "agent",
            "workflow": span.name,
            "start_time": span.start_time,
            "end_time": span.end_time,
        }

        # Add any additional attributes that might be useful
        if "service.name" in span.resource.attributes:
            span_info["service"] = span.resource.attributes["service.name"]

        return span_info

    def _extract_chain_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract information from a CHAIN span."""
        attributes = span.attributes
        if not attributes:
            msg = "Span attributes are empty"
            raise ValueError(msg)

        span_info: dict[str, Any] = {
            "type": "chain",
            "workflow": span.name,
            "start_time": span.start_time,
            "end_time": span.end_time,
        }

        # Extract input and output values
        input_value = attributes.get("input.value", "")
        output_value = attributes.get("output.value", "")

        # Try to parse JSON if available
        try:
            span_info["input"] = json.loads(input_value)
        except (json.JSONDecodeError, TypeError):
            span_info["input"] = input_value

        try:
            span_info["output"] = json.loads(output_value)
        except (json.JSONDecodeError, TypeError):
            span_info["output"] = output_value

        # Add service name if available
        if "service.name" in span.resource.attributes:
            span_info["service"] = span.resource.attributes["service.name"]

        return span_info



from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast
from uuid import uuid4

from any_agent.config import AgentConfig, AgentFramework, TracingConfig
from any_agent.logging import logger

from .any_agent import AnyAgent

try:
    from langchain_core.language_models import LanguageModelLike
    from langchain_litellm import ChatLiteLLM
    from langgraph.prebuilt import create_react_agent
    from langgraph_swarm import create_handoff_tool, create_swarm

    DEFAULT_AGENT_TYPE = create_react_agent
    DEFAULT_MODEL_TYPE = ChatLiteLLM

    langchain_available = True
except ImportError:
    langchain_available = False

if TYPE_CHECKING:
    from langchain_core.language_models import LanguageModelLike
    from langchain_core.messages.base import BaseMessage
    from langgraph.graph.graph import CompiledGraph

    from any_agent.tracing.trace import AgentTrace


class LangchainAgent(AnyAgent):
    """LangChain agent implementation that handles both loading and running."""

    def __init__(
        self,
        config: AgentConfig,
        managed_agents: Sequence[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ):
        super().__init__(config, managed_agents, tracing)
        self._agent: CompiledGraph | None = None
        self._tools: Sequence[Any] = []

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.LANGCHAIN

    def _get_model(self, agent_config: AgentConfig) -> "LanguageModelLike":
        """Get the model configuration for a LangChain agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE

        return cast(
            "LanguageModelLike",
            model_type(
                model=agent_config.model_id,
                api_key=agent_config.api_key,
                api_base=agent_config.api_base,
                model_kwargs=agent_config.model_args or {},  # type: ignore[arg-type]
            ),
        )

    async def _load_agent(self) -> None:
        """Load the LangChain agent with the given configuration."""
        if not langchain_available:
            msg = "You need to `pip install 'any-agent[langchain]'` to use this agent"
            raise ImportError(msg)

        imported_tools, _ = await self._load_tools(self.config.tools)

        if self.managed_agents:
            swarm = []
            managed_names = []
            for n, managed_agent in enumerate(self.managed_agents):
                managed_tools, _ = await self._load_tools(managed_agent.tools)
                name = managed_agent.name
                if not name or name == "any_agent":
                    logger.warning(
                        "Overriding name for managed_agent. Can't use the default.",
                    )
                    name = f"managed_agent_{n}"
                managed_names.append(name)

                agent_type = managed_agent.agent_type or DEFAULT_AGENT_TYPE
                instance = agent_type(
                    name=name,
                    model=self._get_model(managed_agent),
                    tools=[
                        *managed_tools,
                        create_handoff_tool(agent_name=self.config.name),
                    ],
                    prompt=managed_agent.instructions,
                    **managed_agent.agent_args or {},
                )
                swarm.append(instance)

            imported_tools = [
                create_handoff_tool(agent_name=managed_name)
                for managed_name in managed_names
            ]
            agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE
            self._main_agent_tools = imported_tools
            main_agent = agent_type(
                name=self.config.name,
                model=self._get_model(self.config),
                tools=imported_tools,
                prompt=self.config.instructions,
                **self.config.agent_args or {},
            )
            swarm.append(main_agent)
            workflow = create_swarm(swarm, default_active_agent=self.config.name)
            self._agent = workflow.compile()
        else:
            self._main_agent_tools = imported_tools
            agent_type = self.config.agent_type or DEFAULT_AGENT_TYPE
            self._agent = agent_type(
                name=self.config.name,
                model=self._get_model(self.config),
                tools=imported_tools,
                prompt=self.config.instructions,
                **self.config.agent_args or {},
            )
        # Langgraph doesn't let you easily access what tools are loaded from the CompiledGraph,
        # so we'll store a list of them in this class
        self._tools = imported_tools

    async def run_async(self, prompt: str, **kwargs: Any) -> "AgentTrace":
        """Run the LangChain agent with the given prompt."""
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        inputs = {"messages": [("user", prompt)]}
        run_id = str(uuid4())
        tracer = self._tracer_provider.get_tracer("any_agent")
        with tracer.start_as_current_span("agent_run") as span:
            span.set_attribute("any_agent.run_id", run_id)
            result = await self._agent.ainvoke(inputs, **kwargs)
        if not result.get("messages"):
            msg = "No messages returned from the agent."
            raise ValueError(msg)
        last_message: BaseMessage = result["messages"][-1]
        trace = self._exporter.pop_trace(run_id)
        trace.final_output = str(last_message.content)
        return trace



from unittest.mock import MagicMock, patch

from opentelemetry.sdk.trace import ReadableSpan

from any_agent.config import AgentFramework, TracingConfig
from any_agent.tracing.exporter import AnyAgentExporter


def test_exporter_initialization(agent_framework: AgentFramework) -> None:
    exporter = AnyAgentExporter(
        agent_framework=agent_framework,
        tracing_config=TracingConfig(),
    )

    assert exporter.console is not None


def test_rich_console_span_exporter_default(llm_span: ReadableSpan):  # type: ignore[no-untyped-def]
    console_mock = MagicMock()
    with patch("any_agent.tracing.exporter.Console", console_mock):
        exporter = AnyAgentExporter(AgentFramework.LANGCHAIN, TracingConfig())
        exporter.export([llm_span])
        console_mock.return_value.rule.assert_called()


def test_rich_console_span_exporter_disable(llm_span: ReadableSpan):  # type: ignore[no-untyped-def]
    console_mock = MagicMock()
    with patch("any_agent.tracing.exporter.Console", console_mock):
        exporter = AnyAgentExporter(
            AgentFramework.LANGCHAIN,
            TracingConfig(llm=None),
        )
        exporter.export([llm_span])
        console_mock.return_value.rule.assert_not_called()


def test_cost_info_default(llm_span: ReadableSpan):  # type: ignore[no-untyped-def]
    console_mock = MagicMock()
    with patch("any_agent.tracing.exporter.Console", console_mock):
        exporter = AnyAgentExporter(
            AgentFramework.LANGCHAIN,
            TracingConfig(
                console=False,
            ),
        )
        exporter.export([llm_span])
        attributes = (
            exporter.pop_trace(str(llm_span.attributes["any_agent.run_id"]))  # type: ignore[index]
            .spans[0]
            .attributes
        )
        for key in (
            "cost_prompt",
            "cost_completion",
        ):
            assert key in attributes


def test_rich_console_cost_info_disabled(llm_span: ReadableSpan):  # type: ignore[no-untyped-def]
    console_mock = MagicMock()
    with patch("any_agent.tracing.exporter.Console", console_mock):
        exporter = AnyAgentExporter(
            AgentFramework.LANGCHAIN,
            TracingConfig(console=False, cost_info=False),
        )
        exporter.export([llm_span])
        attributes = (
            exporter.pop_trace(str(llm_span.attributes["any_agent.run_id"]))  # type: ignore[index]
            .spans[0]
            .attributes
        )
        for key in (
            "cost_prompt",
            "cost_completion",
        ):
            assert key not in attributes



from __future__ import annotations

import asyncio
import json
from contextlib import suppress
from typing import TYPE_CHECKING, Any

import litellm
from litellm.types.utils import ChatCompletionMessageToolCall

from any_agent.config import AgentConfig, AgentFramework, TracingConfig
from any_agent.logging import logger
from any_agent.tracing.trace import AgentTrace

from .any_agent import AnyAgent

if TYPE_CHECKING:
    from collections.abc import Callable

    from litellm.types.utils import Message as LiteLLMMessage


DEFAULT_SYSTEM_PROMPT = """
You are an agent - please keep going until the user's query is completely resolved, before ending your turn and yielding back to the user. Only terminate your turn when you are sure that the problem is solved, or if you need more info from the user to solve the problem.

If you are not sure about anything pertaining to the user's request, use your tools to read files and gather the relevant information: do NOT guess or make up an answer.

You MUST plan extensively before each function call, and reflect extensively on the outcomes of the previous function calls. DO NOT do this entire process by making function calls only, as this can impair your ability to solve the problem and think insightfully.
""".strip()

# Max number of tool calling + chat completion steps in response to a single user query
DEFAULT_MAX_NUM_TURNS = 10


### Internal tools for tiny-agent ###"
def task_completion_tool() -> dict[str, Any]:
    """Tool to indicate task completion."""
    return {
        "type": "function",
        "function": {
            "name": "task_complete",
            "description": "Call this tool when the task given by the user is complete",
            "parameters": {
                "type": "object",
                "properties": {},
            },
        },
    }


class ToolExecutor:
    """Executor for tools that wraps tool functions to work with the MCP client."""

    def __init__(self, tool_function: Callable[..., Any]) -> None:
        """Initialize the tool executor.

        Args:
            tool_function: The tool function to execute

        """
        self.tool_function = tool_function

    async def call_tool(self, request: dict[str, Any]) -> dict[str, Any]:
        """Call the tool function.

        Args:
            request: The tool request with name and arguments

        Returns:
            Tool execution result

        """
        try:
            # Extract arguments
            arguments = request.get("arguments", {})

            if hasattr(self.tool_function, "__annotations__"):
                func_args = self.tool_function.__annotations__
                for arg_name, arg_type in func_args.items():
                    if arg_name in arguments:
                        with suppress(Exception):
                            # Convert the argument to the expected type
                            arguments[arg_name] = arg_type(arguments[arg_name])

            # Call the tool function
            if asyncio.iscoroutinefunction(self.tool_function):
                result = await self.tool_function(**arguments)
            else:
                result = self.tool_function(**arguments)

            # Format the result
            if isinstance(result, str):
                return {"content": [{"text": result}]}
            if isinstance(result, dict):
                return {"content": [{"text": str(result)}]}
            return {"content": [{"text": str(result)}]}
        except Exception as e:
            return {"content": [{"text": f"Error executing tool: {e}"}]}


class TinyAgent(AnyAgent):
    """A lightweight agent implementation using litellm.

    Modeled after JS implementation https://huggingface.co/blog/tiny-agents.
    """

    def __init__(
        self,
        config: AgentConfig,
        managed_agents: list[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ) -> None:
        """Initialize the TinyAgent.

        Args:
            config: Agent configuration
            managed_agents: Optional list of managed agent configurations
            tracing: Optional tracing configuration

        """
        # we don't yet support multi-agent in tinyagent
        if managed_agents:
            msg = "Managed agents are not supported in TinyAgent."
            raise ValueError(msg)
        super().__init__(config, managed_agents=managed_agents, tracing=tracing)
        self.messages: list[dict[str, Any]] = []
        self.instructions = config.instructions or DEFAULT_SYSTEM_PROMPT
        self.api_key = config.api_key
        self.api_base = config.api_base
        self.model = config.model_id
        self.model_kwargs = config.model_args or {}
        self.clients: dict[str, ToolExecutor] = {}
        self.available_tools: list[dict[str, Any]] = []
        self.exit_loop_tools = [task_completion_tool()]

    async def _load_agent(self) -> None:
        """Load the agent and its tools."""
        # Load tools
        logger.debug("Loading tools: %s", self.config.tools)
        wrapped_tools, mcp_servers = await self._load_tools(self.config.tools)
        self._mcp_servers = (
            mcp_servers  # Store servers so that they don't get garbage collected
        )
        logger.debug("Wrapped tools count: %s", len(wrapped_tools))

        self._main_agent_tools = wrapped_tools

        for tool in wrapped_tools:
            tool_name = tool.__name__
            tool_desc = tool.__doc__ or f"Tool to {tool_name}"

            # check if the tool has __input__schema__ attribute which we set when wrapping MCP tools
            if not hasattr(tool, "__input_schema__"):
                # Generate one from the function signature
                import inspect

                sig = inspect.signature(tool)
                properties = {}
                required = []

                for param_name, param in sig.parameters.items():
                    # Skip *args and **kwargs
                    if param.kind in (
                        inspect.Parameter.VAR_POSITIONAL,
                        inspect.Parameter.VAR_KEYWORD,
                    ):
                        continue

                    # Add the parameter to properties
                    properties[param_name] = {
                        "type": "string",
                        "description": f"Parameter {param_name}",
                    }

                    # If parameter has no default, it's required
                    if param.default == inspect.Parameter.empty:
                        required.append(param_name)

                input_schema = {
                    "type": "object",
                    "properties": properties,
                    "required": required,
                }
            else:
                # Use the provided schema
                input_schema = tool.__input_schema__

            # Add the tool to available tools
            self.available_tools.append(
                {
                    "type": "function",
                    "function": {
                        "name": tool_name,
                        "description": tool_desc,
                        "parameters": input_schema,
                    },
                }
            )

            # Register tool with the client
            self.clients[tool_name] = ToolExecutor(tool)
            logger.debug("Registered tool: %s", tool_name)

    async def run_async(self, prompt: str, **kwargs: Any) -> AgentTrace:
        """Run the agent asynchronously.

        Args:
            prompt: User input prompt
            **kwargs: Additional parameters

        Returns:
            The final agent response

        """
        logger.debug("Running agent with prompt: %s...", prompt[:500])
        max_turns = kwargs.get("max_turns", DEFAULT_MAX_NUM_TURNS)
        self.messages = [
            {
                "role": "system",
                "content": self.instructions or DEFAULT_SYSTEM_PROMPT,
            },
            {
                "role": "user",
                "content": prompt,
            },
        ]

        num_of_turns = 0
        next_turn_should_call_tools = True
        final_response = ""
        assistant_messages = []

        while True:
            try:
                logger.debug("Starting turn %s", num_of_turns + 1)

                last_response = await self._process_single_turn_with_tools(
                    {
                        "exit_loop_tools": self.exit_loop_tools,
                        "exit_if_first_chunk_no_tool": num_of_turns > 0
                        and next_turn_should_call_tools,
                    },
                )

                if last_response:
                    logger.debug(last_response)
                    logger.debug(
                        "Assistant response this turn: %s...",
                        last_response[:50],
                    )
                    assistant_messages.append(last_response)
                    final_response = last_response

            except Exception as err:
                logger.error("Error during turn %s: %s", num_of_turns + 1, err)
                if isinstance(err, Exception) and str(err) == "AbortError":
                    return AgentTrace(final_output=final_response or "Task aborted")
                raise

            num_of_turns += 1
            current_last = self.messages[-1]
            logger.debug("Current role: %s", current_last.get("role"))

            # After a turn, check if we have any content in the last assistant message
            if current_last.get("role") == "assistant" and current_last.get("content"):
                final_response = current_last.get("content", "No content found")
                logger.debug(
                    "Updated final response from assistant message: %s...",
                    final_response[:50],
                )

            # Check exit conditions
            if (
                current_last.get("role") == "tool"
                and current_last.get("name")
                and current_last.get("name")
                in [t["function"]["name"] for t in self.exit_loop_tools]
            ):
                logger.debug(
                    "Exiting because tool %s is an exit tool", current_last.get("name")
                )
                # If task is complete, return the last assistant message before this
                for msg in reversed(self.messages[:-1]):
                    if msg.get("role") == "assistant" and msg.get("content"):
                        return AgentTrace(final_output=msg.get("content"))
                return AgentTrace(final_output=final_response or "Task completed")

            if current_last.get("role") != "tool" and num_of_turns > max_turns:
                logger.debug("Exiting because max turns (%s) reached", max_turns)
                return AgentTrace(
                    final_output=final_response or "Max turns reached",
                )

            if current_last.get("role") != "tool" and next_turn_should_call_tools:
                logger.debug("Exiting because no tools were called when expected")
                return AgentTrace(final_output=final_response or "No tools called")

            if current_last.get("role") == "tool":
                next_turn_should_call_tools = False
                logger.debug("Tool was called, next turn should not call tools")
            else:
                next_turn_should_call_tools = True
                logger.debug("No tool was called, next turn should call tools")

    async def _process_single_turn_with_tools(self, options: dict[str, Any]) -> str:
        """Process a single turn of conversation with potential tool calls.

        Args:
            options: Options including exit_loop_tools, exit_if_first_chunk_no_tool

        Returns:
            The response message or combined tool results

        """
        logger.debug("Start of single turn")

        tools = options["exit_loop_tools"] + self.available_tools

        # Create the completion parameters
        completion_params = {
            "model": self.model,
            "messages": self.messages,
            "tools": tools,
            "tool_choice": "auto",
            **self.model_kwargs,
        }

        # Add API key and base if provided
        if self.api_key:
            completion_params["api_key"] = self.api_key
        if self.api_base:
            completion_params["api_base"] = self.api_base
        logger.debug("Sending new message to LLM: %s", self.messages[-1])
        response = await litellm.acompletion(**completion_params)
        message: LiteLLMMessage = response.choices[0].message
        logger.debug("Response message: %s", message.model_dump())
        # if no tools were called, add the exit tool to the message and return
        if not message.tool_calls and options.get("exit_if_first_chunk_no_tool"):
            logger.debug("No tool calls found in response, adding exit tool")
            message.tool_calls = [
                ChatCompletionMessageToolCall(
                    function=task_completion_tool()["function"],
                )
            ]
        self.messages.append(message.model_dump())

        # Process tool calls if any
        combined_results = []
        exit_tool_called = False

        if message.tool_calls:
            logger.debug(f"Processing {len(message.tool_calls)} tool calls")

            for tool_call in message.tool_calls:
                tool_name = tool_call.function.name
                logger.debug("Processing tool call for: %s", tool_name)
                tool_args = {}

                if tool_call.function.arguments:
                    tool_args = json.loads(tool_call.function.arguments)
                    logger.debug("Tool arguments: %s", tool_args)

                tool_message = {
                    "role": "tool",
                    "tool_call_id": tool_call.id,
                    "content": "",
                    "name": tool_name,
                }

                # Check if tool is in exit loop tools
                exit_tools = options["exit_loop_tools"]
                if exit_tools and tool_name in [
                    t["function"]["name"] for t in exit_tools
                ]:
                    logger.debug("Exit tool called: %s", tool_name)
                    exit_tool_called = True
                    self.messages.append(tool_message)
                    combined_results.append(str(tool_message["content"]))
                    continue

                # Check if the tool exists
                if tool_name not in self.clients:
                    logger.error("Tool %s not found in registered tools", tool_name)
                    tool_message["content"] = (
                        f"Error: No tool found with name: {tool_name}"
                    )
                else:
                    client = self.clients[tool_name]
                    try:
                        logger.debug("Calling tool: %s", tool_name)
                        result = await client.call_tool(
                            {"name": tool_name, "arguments": tool_args}
                        )

                        if (
                            isinstance(result, dict)
                            and "content" in result
                            and isinstance(result["content"], list)
                        ):
                            tool_message["content"] = result["content"][0]["text"]
                        else:
                            tool_message["content"] = str(result)

                        logger.debug(
                            "Tool result: %s...",
                            tool_message["content"][:50]
                            if tool_message["content"]
                            else "Empty",
                        )
                    except Exception as e:
                        logger.error("Error calling tool %s: %s", tool_name, e)
                        tool_message["content"] = f"Error calling tool {tool_name}: {e}"

                self.messages.append(tool_message)
                combined_results.append(str(tool_message["content"]))

            # If an exit tool was called, return early with the combined results
            if exit_tool_called:
                return "\n".join(combined_results)

            return "\n".join(combined_results)
        return str(message.content)

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.TINYAGENT



from any_agent.config import AgentFramework
from any_agent.evaluation.evaluation_case import EvaluationCase
from any_agent.evaluation.evaluators import (
    evaluate_checkpoint,
    evaluate_hypothesis,
    evaluate_qa_squad,
)
from any_agent.evaluation.schemas import TraceEvaluationResult
from any_agent.logging import logger
from any_agent.tracing import TracingProcessor
from any_agent.tracing.trace import AgentTrace


def evaluate(
    evaluation_case: EvaluationCase,
    trace: AgentTrace,
    agent_framework: AgentFramework,
) -> TraceEvaluationResult:
    processor = TracingProcessor.create(agent_framework)
    if not processor:
        msg = f"Processor for {agent_framework} not available."
        raise ValueError(msg)
    hypothesis_answer = trace.final_output
    checkpoint_results = evaluate_checkpoint(
        model=evaluation_case.llm_judge,
        trace=trace,
        checkpoints=evaluation_case.checkpoints,
        processor=processor,
    )
    if hypothesis_answer is not None:
        hypothesis_answer_results = evaluate_hypothesis(
            model=evaluation_case.llm_judge,
            hypothesis_final_output=hypothesis_answer,
            ground_truth_answer_dict=evaluation_case.ground_truth,
            ground_truth_checkpoints=evaluation_case.final_output_criteria,
        )
    else:
        logger.warning(
            "No hypothesis answer found in the trace. Skipping hypothesis evaluation."
        )
        hypothesis_answer_results = []

    if evaluation_case.ground_truth and hypothesis_answer is not None:
        direct_results = evaluate_qa_squad(
            hypothesis_answer=hypothesis_answer,
            ground_truth_answer=evaluation_case.ground_truth,
        )
    else:
        logger.warning(
            "No ground truth answer found in the evaluation case or no hypothesis answer in the trace. Skipping direct evaluation."
        )
        direct_results = []

    return TraceEvaluationResult(
        trace=trace,
        hypothesis_answer=hypothesis_answer,
        checkpoint_results=checkpoint_results,
        hypothesis_answer_results=hypothesis_answer_results,
        direct_results=direct_results,
    )



from collections.abc import Generator, Sequence
from unittest.mock import AsyncMock, patch

import pytest
from agents.mcp import MCPServerSse as OpenAIInternalMCPServerSse

from any_agent.config import AgentConfig, MCPSse, Tool
from any_agent.frameworks.any_agent import AnyAgent


@pytest.fixture
def openai_mcp_sse_server(
    tools: Sequence[Tool],
) -> Generator[OpenAIInternalMCPServerSse]:
    with patch(
        "any_agent.tools.mcp.frameworks.openai.OpenAIInternalMCPServerSse",
    ) as mock_server:
        mock_server.return_value.list_tools = AsyncMock(return_value=tools)
        yield mock_server


@pytest.mark.usefixtures(
    "openai_mcp_sse_server",
)
def test_openai_mcpsse(
    mcp_sse_params_no_tools: MCPSse,
) -> None:
    """This is a test kept for legacy purposes."""
    agent_config = AgentConfig(model_id="gpt-4o", tools=[mcp_sse_params_no_tools])

    agent = AnyAgent.create("openai", agent_config)

    servers = agent._mcp_servers
    assert servers

    server, *_ = agent._mcp_servers
    assert server.mcp_tool == mcp_sse_params_no_tools



from typing import TYPE_CHECKING, Any
from uuid import uuid4

from any_agent.config import AgentConfig, AgentFramework, TracingConfig

from .any_agent import AnyAgent

try:
    from agents import (
        Agent,
        Handoff,
        Model,
        ModelSettings,
        Runner,
    )
    from agents.extensions.models.litellm_model import LitellmModel

    DEFAULT_MODEL_TYPE = LitellmModel

    agents_available = True
except ImportError:
    agents_available = False


if TYPE_CHECKING:
    from agents import Model

    from any_agent.tracing.trace import AgentTrace


class OpenAIAgent(AnyAgent):
    """OpenAI agent implementation that handles both loading and running."""

    def __init__(
        self,
        config: AgentConfig,
        managed_agents: list[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ):
        super().__init__(config, managed_agents, tracing)
        self._agent: Agent | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.OPENAI

    def _get_model(
        self,
        agent_config: AgentConfig,
    ) -> "Model":
        """Get the model configuration for an OpenAI agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE
        return model_type(
            model=agent_config.model_id,
            base_url=agent_config.api_base,
            api_key=agent_config.api_key,
        )

    async def _load_agent(self) -> None:
        """Load the OpenAI agent with the given configuration."""
        if not agents_available:
            msg = "You need to `pip install 'any-agent[openai]'` to use this agent"
            raise ImportError(msg)
        if not agents_available:
            msg = "You need to `pip install openai-agents` to use this agent"
            raise ImportError(msg)

        tools, mcp_servers = await self._load_tools(self.config.tools)
        tools = self._filter_mcp_tools(tools, mcp_servers)

        handoffs = list[Agent[Any] | Handoff[Any]]()
        if self.managed_agents:
            for managed_agent in self.managed_agents:
                managed_tools, managed_mcp_servers = await self._load_tools(
                    managed_agent.tools
                )
                managed_tools = self._filter_mcp_tools(managed_tools, mcp_servers)
                managed_agent_args = managed_agent.agent_args or {}
                handoff = managed_agent_args.pop("handoff", None)
                if managed_agent.model_args:
                    managed_agent_args["model_settings"] = managed_agent.model_args
                instance = Agent(
                    name=managed_agent.name,
                    instructions=managed_agent.instructions,
                    model=self._get_model(managed_agent),
                    tools=managed_tools,
                    mcp_servers=[
                        managed_mcp_server.server
                        for managed_mcp_server in managed_mcp_servers
                    ],
                    **managed_agent_args,
                )
                if handoff:
                    handoffs.append(instance)
                else:
                    tools.append(
                        instance.as_tool(
                            tool_name=instance.name,
                            tool_description=managed_agent.description
                            or f"Use the agent: {managed_agent.name}",
                        ),
                    )

        kwargs_ = self.config.agent_args or {}
        if self.config.model_args:
            kwargs_["model_settings"] = ModelSettings(**self.config.model_args)

        self._main_agent_tools = tools
        self._agent = Agent(
            name=self.config.name,
            instructions=self.config.instructions,
            model=self._get_model(self.config),
            handoffs=handoffs,
            tools=tools,
            mcp_servers=[mcp_server.server for mcp_server in mcp_servers],
            **kwargs_,
        )

    def _filter_mcp_tools(self, tools: list[Any], mcp_servers: list[Any]) -> list[Any]:
        """OpenAI frameowrk doesn't expect the mcp tool to be included in `tools`."""
        non_mcp_tools = []
        for tool in tools:
            if any(tool in mcp_server.tools for mcp_server in mcp_servers):
                continue
            non_mcp_tools.append(tool)
        return non_mcp_tools

    async def run_async(self, prompt: str, **kwargs: Any) -> "AgentTrace":
        """Run the OpenAI agent with the given prompt asynchronously."""
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        tracer = self._tracer_provider.get_tracer("any_agent")
        run_id = str(uuid4())
        with tracer.start_as_current_span("agent_run") as span:
            span.set_attribute("any_agent.run_id", run_id)
            result = await Runner.run(self._agent, prompt, **kwargs)
        trace = self._exporter.pop_trace(run_id)
        trace.final_output = result.final_output
        return trace



import inspect
from collections.abc import Callable, MutableSequence, Sequence
from functools import wraps
from typing import TYPE_CHECKING, Any, TypeVar

from any_agent.config import AgentFramework, MCPParams, Tool
from any_agent.tools import (
    _get_mcp_server,
    _MCPServerBase,
)

if TYPE_CHECKING:
    from agents import Tool as AgentTool
    from google.adk.tools import BaseTool as GoogleTool
    from langchain_core.tools import BaseTool as LangchainTool
    from llama_index.core.tools import FunctionTool as LlamaIndexTool
    from smolagents import Tool as SmolagentsTool


def _wrap_tool_openai(tool: "Tool | AgentTool") -> "AgentTool":
    from agents import Tool as AgentTool
    from agents import function_tool

    if isinstance(tool, AgentTool):  # type: ignore[arg-type, misc]
        return tool  # type: ignore[return-value]

    return function_tool(tool)  # type: ignore[arg-type]


def _wrap_tool_langchain(tool: "Tool | LangchainTool") -> "LangchainTool":
    from langchain_core.tools import BaseTool
    from langchain_core.tools import tool as langchain_tool

    if isinstance(tool, BaseTool):
        return tool

    return langchain_tool(tool)  # type: ignore[arg-type]


def _wrap_tool_smolagents(tool: "Tool | SmolagentsTool") -> "SmolagentsTool":
    from smolagents import Tool as SmolagentsTool
    from smolagents import tool as smolagents_tool

    if isinstance(tool, SmolagentsTool):
        return tool

    # this wrapping needed until https://github.com/huggingface/smolagents/pull/1203 is merged and released
    @wraps(tool)  # type: ignore[arg-type]
    def wrapped_function(*args, **kwargs) -> Any:  # type: ignore[no-untyped-def]
        return tool(*args, **kwargs)  # type: ignore[operator]

    return smolagents_tool(wrapped_function)


def _wrap_tool_llama_index(tool: "Tool | LlamaIndexTool") -> "LlamaIndexTool":
    from llama_index.core.tools import FunctionTool

    if isinstance(tool, FunctionTool):
        return tool

    return FunctionTool.from_defaults(tool)  # type: ignore[arg-type]


def _wrap_tool_google(tool: "Tool | GoogleTool") -> "GoogleTool":
    from google.adk.tools import BaseTool, FunctionTool

    if isinstance(tool, BaseTool):
        return tool

    return FunctionTool(tool)  # type: ignore[arg-type]


def _wrap_tool_agno(tool: Tool) -> Tool:
    # Agno lets you pass callables directly in as tools ❤️
    return tool


def _wrap_tool_tiny(tool: Tool) -> Tool:
    # Tiny lets you pass callables directly in as tools ❤️
    return tool


WRAPPERS: dict[AgentFramework, Callable[..., Any]] = {
    AgentFramework.GOOGLE: _wrap_tool_google,
    AgentFramework.OPENAI: _wrap_tool_openai,
    AgentFramework.LANGCHAIN: _wrap_tool_langchain,
    AgentFramework.SMOLAGENTS: _wrap_tool_smolagents,
    AgentFramework.LLAMA_INDEX: _wrap_tool_llama_index,
    AgentFramework.AGNO: _wrap_tool_agno,
    AgentFramework.TINYAGENT: _wrap_tool_tiny,
}


def verify_callable(tool: Callable[..., Any]) -> None:
    """Verify that `tool` is a valid callable.

    - It needs to have some sort of docstring that describes what it does
    - It needs to have typed argument
    - It needs to have a typed return.

    We need these things because this info gets provided to the agent so that they know how and when to call the tool.
    """
    signature = inspect.signature(tool)
    if not tool.__doc__:
        msg = f"Tool {tool} needs to have a docstring but does not"
        raise ValueError(msg)

    # Check if the function has a return type
    if signature.return_annotation is inspect.Signature.empty:
        msg = f"Tool {tool} needs to have a return type but does not"
        raise ValueError(msg)
    # Check if all parameters have type annotations
    for param in signature.parameters.values():
        if param.annotation is inspect.Signature.empty:
            msg = f"Tool {tool} needs to have typed arguments but does not"
            raise ValueError(msg)


T_co = TypeVar("T_co", covariant=True)


async def _wrap_tools(
    tools: Sequence[T_co],
    agent_framework: AgentFramework,
) -> tuple[list[T_co], list[_MCPServerBase[T_co]]]:
    wrapper = WRAPPERS[agent_framework]

    wrapped_tools = list[T_co]()
    mcp_servers: MutableSequence[_MCPServerBase[T_co]] = []
    for tool in tools:
        # if it's MCPStdio or MCPSse, we need to wrap it in a server
        if isinstance(tool, MCPParams):
            # MCP adapters are usually implemented as context managers.
            # We wrap the server using `MCPServerBase` so the
            # tools can be used as any other callable.
            mcp_server = _get_mcp_server(tool, agent_framework)
            await mcp_server._setup_tools()
            mcp_servers.append(mcp_server)  # type: ignore[arg-type]
        elif callable(tool):
            verify_callable(tool)
            wrapped_tools.append(wrapper(tool))
        else:
            msg = f"Tool {tool} needs to be of type `MCPStdio` or `callable` but is {type(tool)}"
            raise ValueError(msg)

    return wrapped_tools, mcp_servers  # type: ignore[return-value]



from unittest.mock import MagicMock, patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent


def test_load_smolagent_default() -> None:
    mock_agent = MagicMock()
    mock_model = MagicMock()
    mock_tool = MagicMock()

    with (
        patch("any_agent.frameworks.smolagents.DEFAULT_AGENT_TYPE", mock_agent),
        patch("any_agent.frameworks.smolagents.DEFAULT_MODEL_TYPE", mock_model),
        patch("smolagents.tool", mock_tool),
    ):
        AnyAgent.create(
            AgentFramework.SMOLAGENTS,
            AgentConfig(
                model_id="openai/o3-mini",
            ),
        )

        mock_agent.assert_called_once_with(
            name="any_agent",
            model=mock_model.return_value,
            managed_agents=[],
            verbosity_level=-1,
            tools=[],
        )
        mock_model.assert_called_once_with(
            model_id="openai/o3-mini", api_base=None, api_key=None
        )


def test_load_smolagent_with_api_base() -> None:
    mock_agent = MagicMock()
    mock_model = MagicMock()
    mock_tool = MagicMock()

    with (
        patch("any_agent.frameworks.smolagents.DEFAULT_AGENT_TYPE", mock_agent),
        patch("any_agent.frameworks.smolagents.DEFAULT_MODEL_TYPE", mock_model),
        patch("smolagents.tool", mock_tool),
    ):
        AnyAgent.create(
            AgentFramework.SMOLAGENTS,
            AgentConfig(
                model_id="openai/o3-mini",
                model_args={},
                api_base="https://custom-api.example.com",
            ),
        )

        mock_agent.assert_called_once_with(
            name="any_agent",
            model=mock_model.return_value,
            managed_agents=[],
            tools=[],
            verbosity_level=-1,
        )
        mock_model.assert_called_once_with(
            model_id="openai/o3-mini",
            api_base="https://custom-api.example.com",
            api_key=None,
        )


def test_load_smolagents_agent_missing() -> None:
    with patch("any_agent.frameworks.smolagents.smolagents_available", False):
        with pytest.raises(ImportError):
            AnyAgent.create(AgentFramework.SMOLAGENTS, AgentConfig(model_id="gpt-4o"))


def test_run_smolagent_custom_args() -> None:
    mock_agent = MagicMock()
    mock_agent.return_value = MagicMock()
    with (
        patch("any_agent.frameworks.smolagents.DEFAULT_AGENT_TYPE", mock_agent),
        patch("any_agent.frameworks.smolagents.DEFAULT_MODEL_TYPE"),
        patch("smolagents.tool"),
    ):
        agent = AnyAgent.create(
            AgentFramework.SMOLAGENTS,
            AgentConfig(
                model_id="openai/o3-mini",
            ),
        )
        agent.run("foo", max_steps=30)
        mock_agent.return_value.run.assert_called_once_with("foo", max_steps=30)



from collections.abc import Generator, Sequence
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from llama_index.tools.mcp import BasicMCPClient as LlamaIndexMCPClient
from llama_index.tools.mcp import McpToolSpec as LlamaIndexMcpToolSpec

from any_agent.config import AgentFramework, MCPSse, Tool
from any_agent.tools import _get_mcp_server


@pytest.fixture
def llama_index_mcp_client() -> Generator[LlamaIndexMCPClient]:
    with patch(
        "any_agent.tools.mcp.frameworks.llama_index.LlamaIndexMCPClient"
    ) as mock_client:
        yield mock_client


@pytest.fixture
def llama_index_mcp_tool_spec(
    tools: Sequence[Tool],
) -> Generator[LlamaIndexMcpToolSpec]:
    tool_spec = MagicMock()
    tool_spec.to_tool_list_async = AsyncMock(return_value=tools)
    with patch(
        "any_agent.tools.mcp.frameworks.llama_index.LlamaIndexMcpToolSpec",
        return_value=tool_spec,
    ) as mock_tool_spec:
        yield mock_tool_spec


@pytest.mark.asyncio
@pytest.mark.usefixtures(
    "llama_index_mcp_tool_spec",
)
async def test_llamaindex_mcp_sse_integration(
    mcp_sse_params_with_tools: MCPSse,
    llama_index_mcp_client: LlamaIndexMCPClient,
) -> None:
    server = _get_mcp_server(mcp_sse_params_with_tools, AgentFramework.LLAMA_INDEX)

    await server._setup_tools()

    llama_index_mcp_client.assert_called_once_with(  # type: ignore[attr-defined]
        command_or_url=mcp_sse_params_with_tools.url
    )



import pytest
from opentelemetry.sdk.trace import ReadableSpan

from any_agent import AgentFramework
from any_agent.tracing import TracingProcessor
from any_agent.tracing.trace import AgentSpan, _is_tracing_supported


def test_extract_interaction(
    agent_framework: AgentFramework, llm_span: ReadableSpan
) -> None:
    if not _is_tracing_supported(agent_framework):
        pytest.skip()
    processor = TracingProcessor.create(AgentFramework(agent_framework))
    # to make mypy happy
    assert processor
    assert llm_span.attributes

    span = AgentSpan.from_readable_span(llm_span)
    span_kind, interaction = processor.extract_interaction(span)
    assert span_kind == "LLM"
    assert interaction["input"]



import asyncio
import os
import subprocess
from datetime import datetime
from pathlib import Path
from typing import Any

import pytest
from litellm.utils import validate_environment

from any_agent import AgentConfig, AgentFramework, AnyAgent, TracingConfig
from any_agent.config import MCPStdio
from any_agent.evaluation import EvaluationCase, evaluate
from any_agent.evaluation.schemas import CheckpointCriteria, TraceEvaluationResult
from any_agent.tracing.trace import AgentTrace, _is_tracing_supported


def check_uvx_installed() -> bool:
    """The integration tests requires uvx"""
    try:
        result = subprocess.run(  # noqa: S603
            ["uvx", "--version"],  # noqa: S607
            capture_output=True,
            check=True,
        )
        return True if result.returncode == 0 else False  # noqa: TRY300
    except Exception:
        return False


@pytest.mark.skipif(
    os.environ.get("ANY_AGENT_INTEGRATION_TESTS", "FALSE").upper() != "TRUE",
    reason="Integration tests require `ANY_AGENT_INTEGRATION_TESTS=TRUE` env var",
)
def test_load_and_run_agent(agent_framework: AgentFramework, tmp_path: Path) -> None:
    kwargs = {}

    tmp_file = "tmp.txt"

    if not check_uvx_installed():
        msg = "uvx is not installed. Please install it to run this test."
        raise RuntimeError(msg)

    def write_file(text: str) -> None:
        """write the text to a file in the tmp_path directory

        Args:
            text (str): The text to write to the file.

        Returns:
            None
        """
        with open(os.path.join(tmp_path, tmp_file), "w", encoding="utf-8") as f:
            f.write(text)

    kwargs["model_id"] = "gpt-4.1-mini"
    env_check = validate_environment(kwargs["model_id"])
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args: dict[str, Any] = (
        {"parallel_tool_calls": False}
        if agent_framework not in [AgentFramework.AGNO, AgentFramework.LLAMA_INDEX]
        else {}
    )
    model_args["temperature"] = 0.0
    tools = [
        write_file,
        MCPStdio(
            command="uvx",
            args=["mcp-server-time", "--local-timezone=America/New_York"],
            tools=[
                "get_current_time",
            ],
        ),
    ]
    agent_config = AgentConfig(
        tools=tools,  # type: ignore[arg-type]
        instructions="Search the web to answer",
        model_args=model_args,
        **kwargs,  # type: ignore[arg-type]
    )
    agent = AnyAgent.create(agent_framework, agent_config, tracing=TracingConfig())

    try:
        agent_trace = agent.run(
            "Use the tools to find what year it is in the America/New_York timezone and write the value (single number) to a file",
        )
        assert os.path.exists(os.path.join(tmp_path, tmp_file))
        with open(os.path.join(tmp_path, tmp_file)) as f:
            content = f.read()
        assert content == str(datetime.now().year)
        assert isinstance(agent_trace, AgentTrace)
        assert agent_trace.final_output
        if _is_tracing_supported(agent_framework):
            assert agent_trace.spans
            assert len(agent_trace.spans) > 0
            cost_sum = agent_trace.get_total_cost()
            assert cost_sum.total_cost > 0
            assert cost_sum.total_cost < 1.00
            assert cost_sum.total_tokens > 0
            assert cost_sum.total_tokens < 20000
            case = EvaluationCase(
                llm_judge="gpt-4.1-mini",
                checkpoints=[
                    CheckpointCriteria(
                        criteria="Check if the agent called the write_file tool and it succeeded",
                        points=1,
                    ),
                    CheckpointCriteria(
                        criteria="Check if the agent wrote the year to the file.",
                        points=1,
                    ),
                    CheckpointCriteria(
                        criteria="Check if the year was 1990",
                        points=1,
                    ),
                ],
            )
            result: TraceEvaluationResult = evaluate(
                evaluation_case=case,
                trace=agent_trace,
                agent_framework=agent_framework,
            )
            assert result
            assert result.score == float(2 / 3)
    finally:
        agent.exit()


@pytest.mark.asyncio
async def test_run_agent_twice(agent_framework: AgentFramework) -> None:
    """When an agent is run twice, state from the first run shouldn't bleed into the second run"""
    model_id = "gpt-4.1-nano"
    env_check = validate_environment(model_id)
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args = {"temperature": 0.0}
    try:
        agent = await AnyAgent.create_async(
            agent_framework,
            AgentConfig(model_id=model_id, model_args=model_args, tools=[]),
        )
        results = await asyncio.gather(
            agent.run_async("What is the capital of France?"),
            agent.run_async("What is the capital of Spain?"),
        )
        result1, result2 = results
        assert result1.final_output is not None
        assert result2.final_output is not None
        assert "Paris" in result1.final_output
        assert "Madrid" in result2.final_output
        if _is_tracing_supported(agent_framework):
            first_spans = result1.spans
            second_spans = result2.spans
            assert second_spans[: len(first_spans)] != first_spans, (
                "Spans from the first run should not be in the second"
            )
            assert result1.spans
            assert len(result1.spans) > 0
            cost_sum = result1.get_total_cost()
            assert cost_sum.total_cost > 0
            assert cost_sum.total_cost < 1.00
            assert cost_sum.total_tokens > 0
            assert cost_sum.total_tokens < 20000
    finally:
        agent.exit()



from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.tools import (
    search_web,
    visit_webpage,
)


def test_load_agno_default() -> None:
    mock_agent = MagicMock()
    mock_model = MagicMock()

    with (
        patch("any_agent.frameworks.agno.Agent", mock_agent),
        patch("any_agent.frameworks.agno.DEFAULT_MODEL_TYPE", mock_model),
    ):
        AnyAgent.create(AgentFramework.AGNO, AgentConfig(model_id="gpt-4o"))
        mock_agent.assert_called_once_with(
            name="any_agent",
            instructions=None,
            model=mock_model(model="gpt-4o"),
            tools=[],
        )


def test_load_agno_agent_missing() -> None:
    with patch("any_agent.frameworks.agno.agno_available", False):
        with pytest.raises(ImportError):
            AnyAgent.create(AgentFramework.AGNO, AgentConfig(model_id="gpt-4o"))


def test_load_agno_multi_agent() -> None:
    mock_agent = MagicMock()
    mock_team = MagicMock()
    mock_model = MagicMock()

    with (
        patch("any_agent.frameworks.agno.Agent", mock_agent),
        patch("any_agent.frameworks.agno.Team", mock_team),
        patch("any_agent.frameworks.agno.DEFAULT_MODEL_TYPE", mock_model),
    ):
        AnyAgent.create(
            AgentFramework.AGNO,
            AgentConfig(model_id="gpt-4o", tools=[search_web]),
            managed_agents=[
                AgentConfig(
                    model_id="gpt-4o-mini",
                    name="search-web-agent",
                    description="You can visit webpages",
                    tools=[
                        visit_webpage,
                    ],
                )
            ],
        )
        mock_agent.assert_called_once_with(
            name="search-web-agent",
            role="You can visit webpages",
            instructions=None,
            model=mock_model(model="gpt-4o-mini"),
            tools=[visit_webpage],
        )
        mock_team.assert_called_once_with(
            mode="collaborate",
            name="Team managed by agent any_agent",
            description=None,
            instructions=None,
            model=mock_model(model="gpt-4o"),
            members=[mock_agent.return_value],
            tools=[search_web],
        )


def test_run_agno_custom_args() -> None:
    mock_agent = MagicMock()
    # Create a mock response object with the required content attribute
    mock_response = MagicMock()
    mock_response.content = "mock response"

    # Set up the AsyncMock to return the mock response
    mock_agent_instance = AsyncMock()
    mock_agent_instance.arun.return_value = mock_response
    mock_agent.return_value = mock_agent_instance

    mock_model = MagicMock()

    with (
        patch("any_agent.frameworks.agno.Agent", mock_agent),
        patch("any_agent.frameworks.agno.DEFAULT_MODEL_TYPE", mock_model),
    ):
        agent = AnyAgent.create(AgentFramework.AGNO, AgentConfig(model_id="gpt-4o"))
        result = agent.run("foo", retries=2)

        # Verify the result is as expected
        assert isinstance(result.final_output, str)
        assert result.final_output == "mock response"

        # Verify the agent was called with the right parameters
        mock_agent_instance.arun.assert_called_once_with("foo", retries=2)



import logging
import os
from collections.abc import Callable

import pytest
from litellm.utils import validate_environment
from rich.logging import RichHandler

from any_agent import AgentConfig, AgentFramework, AnyAgent
from any_agent.config import TracingConfig
from any_agent.tools import search_web, visit_webpage
from any_agent.tracing.trace import AgentSpan, AgentTrace, _is_tracing_supported

FORMAT = "%(message)s"
logging.basicConfig(
    level=logging.DEBUG,
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)],
)
logger = logging.getLogger("any_agent_test")
logger.setLevel(logging.DEBUG)


CHILD_TAG = "any_agent.children"


def organize(items: list[AgentSpan]) -> None:
    traces = {}
    for trace in items:
        k = trace.context.span_id
        trace.attributes[CHILD_TAG] = {}
        traces[k] = trace
    for trace in items:
        if trace.parent:
            parent_k = trace.parent.span_id
            if parent_k:
                traces[parent_k].attributes[CHILD_TAG][trace.context.span_id] = trace
            else:
                traces[None] = trace
    logger.info(traces[None].model_dump_json(indent=2))


@pytest.mark.skipif(
    os.environ.get("ANY_AGENT_INTEGRATION_TESTS", "FALSE").upper() != "TRUE",
    reason="Integration tests require `ANY_AGENT_INTEGRATION_TESTS=TRUE` env var",
)
def test_load_and_run_multi_agent(
    agent_framework: AgentFramework,
    check_multi_tool_usage: Callable[[list[AgentSpan]], None],
) -> None:
    kwargs = {}

    if agent_framework is AgentFramework.TINYAGENT:
        pytest.skip(
            f"Skipping test for {agent_framework.name} because it does not support multi-agent"
        )

    kwargs["model_id"] = "gpt-4.1-nano"
    agent_model = kwargs["model_id"]
    env_check = validate_environment(kwargs["model_id"])
    if not env_check["keys_in_environment"]:
        pytest.skip(f"{env_check['missing_keys']} needed for {agent_framework}")

    model_args = (
        {"parallel_tool_calls": False}
        if agent_framework not in [AgentFramework.AGNO, AgentFramework.LLAMA_INDEX]
        else None
    )
    main_agent = AgentConfig(
        instructions="Use the available tools to complete the task to obtain additional information to answer the query.",
        description="The orchestrator that can use other agents.",
        model_args=model_args,
        **kwargs,  # type: ignore[arg-type]
    )

    managed_agents = [
        AgentConfig(
            name="search_web_agent",
            model_id=agent_model,
            description="Agent that can search the web. It can find answers on the web if the query cannot be answered.",
            tools=[search_web],
            model_args=model_args,
        ),
        AgentConfig(
            name="visit_webpage_agent",
            model_id=agent_model,
            description="Agent that can visit webpages",
            tools=[visit_webpage],
            model_args=model_args,
        ),
    ]
    agent = AnyAgent.create(
        agent_framework=agent_framework,
        agent_config=main_agent,
        managed_agents=managed_agents,
        tracing=TracingConfig(console=False, cost_info=True),
    )

    try:
        agent_trace = agent.run(
            "Which LLM agent framework is the most appropriate to execute SQL queries using grammar constrained decoding? I am working on a business environment on my own premises, and I would prefer hosting an open source model myself."
        )

        assert isinstance(agent_trace, AgentTrace)
        assert agent_trace.final_output
        if _is_tracing_supported(agent_framework):
            assert agent_trace.spans
            assert len(agent_trace.spans) > 0
            cost_sum = agent_trace.get_total_cost()
            assert cost_sum.total_cost > 0
            assert cost_sum.total_cost < 1.00
            assert cost_sum.total_tokens > 0
            assert cost_sum.total_tokens < 20000
            traces = agent_trace.spans
            organize(traces)
            if agent_framework == AgentFramework.AGNO:
                check_multi_tool_usage(traces)
            else:
                logger.warning(
                    "See https://github.com/mozilla-ai/any-agent/issues/256, multi-agent trace checks not working"
                )
    finally:
        agent.exit()



import json
from typing import TYPE_CHECKING, Any

from any_agent import AgentFramework
from any_agent.tracing import TracingProcessor
from any_agent.tracing.otel_types import StatusCode

if TYPE_CHECKING:
    from any_agent.tracing.trace import AgentSpan


class LangchainTracingProcessor(TracingProcessor):
    """Processor for Langchain agent trace data."""

    def _get_agent_framework(self) -> AgentFramework:
        return AgentFramework.LANGCHAIN

    def _extract_llm_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract LLM interaction details from a span."""
        attributes = span.attributes
        span_info = {
            "model": attributes.get("llm.model_name", "Unknown model"),
            "type": "reasoning",
        }

        if "llm.input_messages.0.message.content" in attributes:
            span_info["input"] = attributes["llm.input_messages.0.message.content"]

        if "llm.output_messages.0.message.content" in attributes:
            span_info["output"] = attributes["llm.output_messages.0.message.content"]

        return span_info

    def _extract_tool_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract tool interaction details from a span."""
        attributes = span.attributes
        tool_info: dict[str, Any] = {
            "tool_name": attributes.get("tool.name", span.name),
            "status": "success"
            if span.status.status_code is StatusCode.OK
            else "error",
            "error": span.status.description,
        }

        if "input.value" in attributes:
            try:
                input_value = json.loads(attributes["input.value"])
                tool_info["input"] = input_value
            except Exception:
                tool_info["input"] = attributes["input.value"]

        if "output.value" in attributes:
            try:
                output_value = json.loads(attributes["output.value"])
                if "output" in output_value:
                    parsed_output = self.parse_generic_key_value_string(
                        output_value["output"],
                    )
                    tool_info["output"] = parsed_output.get("content", parsed_output)
                else:
                    tool_info["output"] = output_value
            except Exception:
                tool_info["output"] = attributes["output.value"]

        return tool_info

    def _extract_chain_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract chain interaction details from a span."""
        attributes = span.attributes
        chain_info: dict[str, Any] = {"type": "chain", "name": span.name}

        # Extract input from the chain
        if "input.value" in attributes:
            try:
                input_data = json.loads(attributes["input.value"])
                if "messages" in input_data and isinstance(
                    input_data["messages"],
                    list,
                ):
                    # Extract message content if available
                    messages = []
                    for msg in input_data["messages"]:
                        if isinstance(msg, list) and len(msg) > 1:
                            # Format: ["role", "content"]
                            messages.append({"role": msg[0], "content": msg[1]})
                    if messages:
                        chain_info["input"] = messages
                    else:
                        chain_info["input"] = input_data
                else:
                    chain_info["input"] = input_data
            except Exception:
                chain_info["input"] = attributes["input.value"]

        # Extract output from the chain
        if "output.value" in attributes:
            try:
                output_data = json.loads(attributes["output.value"])
                if "messages" in output_data:
                    # Try to parse the messages
                    parsed_messages = []
                    for msg in output_data["messages"]:
                        parsed_msg = self.parse_generic_key_value_string(msg)
                        if parsed_msg:
                            parsed_messages.append(parsed_msg)

                    if parsed_messages:
                        chain_info["output"] = parsed_messages
                    else:
                        chain_info["output"] = output_data["messages"]
                else:
                    chain_info["output"] = output_data
            except Exception:
                chain_info["output"] = attributes["output.value"]

        return chain_info

    def _extract_agent_interaction(self, span: "AgentSpan") -> dict[str, Any]:
        """Extract agent interaction details from a span."""
        attributes = span.attributes
        agent_info: dict[str, Any] = {"type": "agent", "name": span.name}

        # Extract input from the agent span
        if "input.value" in attributes:
            try:
                input_data = json.loads(attributes["input.value"])
                if "messages" in input_data and isinstance(
                    input_data["messages"],
                    list,
                ):
                    if len(input_data["messages"]) > 0:
                        # For Langchain agents, messages might be serialized as strings
                        message = self.parse_generic_key_value_string(
                            input_data["messages"][0],
                        )
                        if message and "content" in message:
                            agent_info["input"] = message["content"]
                        else:
                            agent_info["input"] = input_data["messages"][0]
                    else:
                        agent_info["input"] = input_data
                else:
                    agent_info["input"] = input_data
            except Exception:
                agent_info["input"] = attributes["input.value"]

        # Extract direct LLM input if available
        if "llm.input_messages.0.message.content" in attributes:
            agent_info["query"] = attributes["llm.input_messages.0.message.content"]

        # Extract output from the agent span
        if "output.value" in attributes:
            try:
                output_data = json.loads(attributes["output.value"])
                if "messages" in output_data and isinstance(
                    output_data["messages"],
                    list,
                ):
                    if len(output_data["messages"]) > 0:
                        # For Langchain agents, messages might be serialized as strings
                        message = self.parse_generic_key_value_string(
                            output_data["messages"][0],
                        )
                        if message and "content" in message:
                            agent_info["output"] = message["content"]
                        else:
                            agent_info["output"] = output_data["messages"][0]
                    else:
                        agent_info["output"] = output_data
                else:
                    agent_info["output"] = output_data
            except Exception:
                agent_info["output"] = attributes["output.value"]

        # Extract metadata if available
        if "metadata" in attributes:
            try:
                metadata = json.loads(attributes["metadata"])
                agent_info["metadata"] = metadata
            except Exception:
                agent_info["metadata"] = attributes["metadata"]

        return agent_info



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, assert_never

from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor

from any_agent.config import (
    AgentConfig,
    AgentFramework,
    ServingConfig,
    Tool,
    TracingConfig,
)
from any_agent.logging import logger
from any_agent.tools.wrappers import _wrap_tools
from any_agent.tracing.exporter import (
    AnyAgentExporter,
    Instrumenter,
    get_instrumenter_by_framework,
)
from any_agent.tracing.trace import _is_tracing_supported

if TYPE_CHECKING:
    from collections.abc import Sequence

    from any_agent.tools.mcp.mcp_server import _MCPServerBase
    from any_agent.tracing.trace import AgentTrace


class AnyAgent(ABC):
    """Base abstract class for all agent implementations.

    This provides a unified interface for different agent frameworks.
    """

    def __init__(
        self,
        config: AgentConfig,
        managed_agents: Sequence[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ):
        self.config = config
        self.managed_agents = managed_agents

        self._mcp_servers: list[_MCPServerBase[Any]] = []
        self._main_agent_tools: list[Any] = []

        # Tracing is enabled by default
        self._tracing_config: TracingConfig = tracing or TracingConfig()
        self._instrumenter: Instrumenter | None = None
        self._setup_tracing()

    @staticmethod
    def _get_agent_type_by_framework(
        framework_raw: AgentFramework | str,
    ) -> type[AnyAgent]:
        framework = AgentFramework.from_string(framework_raw)

        if framework is AgentFramework.SMOLAGENTS:
            from any_agent.frameworks.smolagents import SmolagentsAgent

            return SmolagentsAgent

        if framework is AgentFramework.LANGCHAIN:
            from any_agent.frameworks.langchain import LangchainAgent

            return LangchainAgent

        if framework is AgentFramework.OPENAI:
            from any_agent.frameworks.openai import OpenAIAgent

            return OpenAIAgent

        if framework is AgentFramework.LLAMA_INDEX:
            from any_agent.frameworks.llama_index import LlamaIndexAgent

            return LlamaIndexAgent

        if framework is AgentFramework.GOOGLE:
            from any_agent.frameworks.google import GoogleAgent

            return GoogleAgent

        if framework is AgentFramework.AGNO:
            from any_agent.frameworks.agno import AgnoAgent

            return AgnoAgent

        if framework is AgentFramework.TINYAGENT:
            from any_agent.frameworks.tinyagent import TinyAgent

            return TinyAgent

        assert_never(framework)

    @classmethod
    def create(
        cls,
        agent_framework: AgentFramework | str,
        agent_config: AgentConfig,
        managed_agents: list[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ) -> AnyAgent:
        """Create an agent using the given framework and config."""
        return asyncio.get_event_loop().run_until_complete(
            cls.create_async(
                agent_framework=agent_framework,
                agent_config=agent_config,
                managed_agents=managed_agents,
                tracing=tracing,
            )
        )

    @classmethod
    async def create_async(
        cls,
        agent_framework: AgentFramework | str,
        agent_config: AgentConfig,
        managed_agents: list[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ) -> AnyAgent:
        """Create an agent using the given framework and config."""
        agent_cls = cls._get_agent_type_by_framework(agent_framework)
        agent = agent_cls(agent_config, managed_agents=managed_agents, tracing=tracing)
        await agent._load_agent()
        return agent

    async def _load_tools(
        self, tools: Sequence[Tool]
    ) -> tuple[list[Any], list[_MCPServerBase[Any]]]:
        tools, mcp_servers = await _wrap_tools(tools, self.framework)
        # Add to agent so that it doesn't get garbage collected
        self._mcp_servers.extend(mcp_servers)
        for mcp_server in mcp_servers:
            tools.extend(mcp_server.tools)
        return tools, mcp_servers

    def _setup_tracing(self) -> None:
        """Initialize the tracing for the agent."""
        self._tracer_provider = TracerProvider()
        trace.set_tracer_provider(self._tracer_provider)
        if not _is_tracing_supported(self.framework):
            logger.warning("Tracing is not yet supported for %s", self.framework.name)
            self._instrumenter = None
            return
        self._exporter = AnyAgentExporter(self.framework, self._tracing_config)
        self._tracer_provider.add_span_processor(SimpleSpanProcessor(self._exporter))
        self._instrumenter = get_instrumenter_by_framework(self.framework)
        self._instrumenter.instrument(tracer_provider=self._tracer_provider)

    def run(self, prompt: str, **kwargs: Any) -> AgentTrace:
        """Run the agent with the given prompt."""
        return asyncio.get_event_loop().run_until_complete(
            self.run_async(prompt, **kwargs)
        )

    def serve(self, serving_config: ServingConfig | None = None) -> None:
        """Serve this agent using the Agent2Agent Protocol (A2A).

        Args:
            serving_config: See [ServingConfig][any_agent.config.ServingConfig].

        Raises:
            ImportError: If the `serving` dependencies are not installed.

        """
        try:
            from any_agent.serving import _get_a2a_server
        except ImportError as e:
            msg = "You need to `pip install 'git+https://github.com/google/A2A#subdirectory=samples/python' to use this method."
            raise ImportError(msg) from e

        server = _get_a2a_server(self, serving_config=serving_config or ServingConfig())
        server.start()

    @abstractmethod
    async def _load_agent(self) -> None:
        """Load the agent instance."""

    @abstractmethod
    async def run_async(self, prompt: str, **kwargs: Any) -> AgentTrace:
        """Run the agent asynchronously with the given prompt."""

    @property
    @abstractmethod
    def framework(self) -> AgentFramework:
        """The Agent Framework used."""

    @property
    def agent(self) -> Any:
        """The underlying agent implementation from the framework.

        This property is intentionally restricted to maintain framework abstraction
        and prevent direct dependency on specific agent implementations.

        If you need functionality that relies on accessing the underlying agent:
        1. Consider if the functionality can be added to the AnyAgent interface
        2. Submit a GitHub issue describing your use case
        3. Contribute a PR implementing the needed functionality

        Raises:
            NotImplementedError: Always raised when this property is accessed

        """
        msg = "Cannot access the 'agent' property of AnyAgent, if you need to use functionality that relies on the underlying agent framework, please file a Github Issue or we welcome a PR to add the functionality to the AnyAgent class"
        raise NotImplementedError(msg)

    def exit(self) -> None:
        """Exit the agent and clean up resources."""
        if self._instrumenter is not None:
            self._instrumenter.uninstrument()  # otherwise, this gets called in the __del__ method of Tracer
            self._instrumenter = None
        self._mcp_servers = []  # drop references to mcp servers so that they get garbage collected



from typing import TYPE_CHECKING, Any

from any_agent.config import AgentConfig, AgentFramework, TracingConfig
from any_agent.logging import logger
from any_agent.tracing.trace import AgentTrace

from .any_agent import AnyAgent

try:
    from agno.agent import Agent
    from agno.models.litellm import LiteLLM
    from agno.team.team import Team
    from agno.tools.toolkit import Toolkit

    DEFAULT_MODEL_TYPE = LiteLLM
    agno_available = True
except ImportError:
    agno_available = False


if TYPE_CHECKING:
    from agno.agent import RunResponse
    from agno.models.base import Model


class AgnoAgent(AnyAgent):
    """Agno agent implementation that handles both loading and running."""

    def __init__(
        self,
        config: AgentConfig,
        managed_agents: list[AgentConfig] | None = None,
        tracing: TracingConfig | None = None,
    ):
        super().__init__(config, managed_agents, tracing)
        self._agent: Agent | Team | None = None

    @property
    def framework(self) -> AgentFramework:
        return AgentFramework.AGNO

    def _get_model(self, agent_config: AgentConfig) -> "Model":
        """Get the model configuration for an Agno agent."""
        model_type = agent_config.model_type or DEFAULT_MODEL_TYPE

        return model_type(
            id=agent_config.model_id,
            api_base=agent_config.api_base,
            api_key=agent_config.api_key,
            request_params=agent_config.model_args or {},  # type: ignore[arg-type]
        )

    @staticmethod
    def _unpack_tools(tools: list[Any]) -> list[Any]:
        unpacked: list[Any] = []
        for tool in tools:
            if isinstance(tool, Toolkit):
                unpacked.extend(f for f in tool.functions.values())
            else:
                unpacked.append(tool)
        return unpacked

    async def _load_agent(self) -> None:
        if not agno_available:
            msg = "You need to `pip install 'any-agent[agno]'` to use this agent"
            raise ImportError(msg)

        if self.managed_agents:
            members = []
            for n, managed_agent in enumerate(self.managed_agents):
                managed_tools, _ = await self._load_tools(managed_agent.tools)
                name = managed_agent.name
                if not name or name == "any_agent":
                    logger.warning(
                        "Overriding name for managed_agent. Can't use the default.",
                    )
                    name = f"managed_agent_{n}"
                members.append(
                    Agent(
                        name=name,
                        role=managed_agent.description,
                        instructions=managed_agent.instructions,
                        model=self._get_model(managed_agent),
                        tools=managed_tools,
                        **managed_agent.agent_args or {},
                    )
                )

            tools, _ = await self._load_tools(self.config.tools)
            self._main_agent_tools = self._unpack_tools(tools)

            self._agent = Team(
                mode="collaborate",
                name=f"Team managed by agent {self.config.name}",
                description=self.config.description,
                model=self._get_model(self.config),
                members=members,  # type: ignore[arg-type]
                instructions=self.config.instructions,
                tools=tools,
                **self.config.agent_args or {},
            )
        else:
            tools, _ = await self._load_tools(self.config.tools)

            self._main_agent_tools = self._unpack_tools(tools)
            self._agent = Agent(
                name=self.config.name,
                instructions=self.config.instructions,
                model=self._get_model(self.config),
                tools=tools,
                **self.config.agent_args or {},
            )

    async def run_async(self, prompt: str, **kwargs: Any) -> "AgentTrace":
        if not self._agent:
            error_message = "Agent not loaded. Call load_agent() first."
            raise ValueError(error_message)
        result: RunResponse = await self._agent.arun(prompt, **kwargs)
        return AgentTrace(
            final_output=result.content,
        )
